/*
  Copyright (C) 2002-2005 JAsCo Development Team
  http://ssel.vub.ac.be/jasco

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
 */
package jasco;


import jasco.options.Options;
import jasco.runtime.connector.Connector;
import jasco.tools.BeanTransformerJavassist;
import jasco.tools.jascoparser.JascoParserTokenManager;
import jasco.tools.jascoparser.JavaCharStream;
import jasco.util.generators.JavaGenerator;
import jasco.util.logging.ExceptionLogger;
import jasco.util.logging.Logger;
import jasco.util.logging.OutputLogger;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileWriter;
import java.io.IOException;
import java.lang.reflect.Method;
import java.util.Iterator;
import java.util.StringTokenizer;
import java.util.Vector;
/**
 * JAsCo facade class for compilation and loading of aspect beans, connectors and traversal connectors. Can also be used
 * to insert traps into java beans.
 * <p>
 * Use setExceptionLogger when you want to get exceptions instead of errors on the error stream when something goes wrong!
 *
 * @see jasco.options.Options
 */

 
public class Jasco  {
	
	
	/**
	 * Invokes a traversal connector 
	 *
     * @param traversalName		fully qualified classname of a traversal connector
	 * @param startingObject	the starting object for the traversal
	 * @param resultMethod	the method for fetching the result, if null no result is fetched
	 * @return	an iterator to all aspect beans in the traversal for fetching results
     */
	public static Object invokeTraversal(String traversalName, Object startingObject, String resultMethod)  {
		
		try {
		Class cl = Class.forName(traversalName);
		Method m = cl.getMethod("getTraversal",new Class[0]);
		Object o = m.invoke(null,new Object[0]);
		Method m2 = cl.getMethod("traverse",new Class[] {java.lang.Object.class});
		
		m2.invoke(o,new Object[] {startingObject});
		
		if(resultMethod==null) 
			return null;
			
		Method m3 = cl.getMethod("getHooks",new Class[0]);
		Iterator i = (Iterator) m3.invoke(o,new Object[0]);
		while (i.hasNext())  {
			try {
			Object hook = i.next();
			Class cl2 = hook.getClass();
			Method m4 = null;
			m4 = cl2.getMethod(resultMethod,new Class[0]);
			return m4.invoke(hook,new Object[0]);
			}
			catch(Throwable ex)  {
				throw new IllegalArgumentException("Method "+resultMethod+" not found in hook (must be public!)");
			}
		}
		} 
		catch(Exception ex)  {
			System.err.println("Exception occured: "+ex.getMessage());
			ex.printStackTrace();
		}
		return null;
		 
	}
	
	public static Object invokeTraversal(String traversalName, Object startingObject)  {
		return invokeTraversal(traversalName,startingObject,null);
	}
	
	
	/**
	 * Makes sure JAsCo throws an exception instead of printing on the error stream when there are errors. 
	 *
     */
	public static void setExceptionLogger()  {
		jasco.util.logging.Logger.setInstance(new ExceptionLogger());
	}
	
	/**
	 * Makes sure JAsCo prints on the error stream when there are errors. 
	 *
     */
	public static void setOutputLogger()  {
		jasco.util.logging.Logger.setInstance(new OutputLogger());
	}
	
	/**
     * sets the output directory where all entities generated by this class will be placed 
	 *
     * @param outputdir   a string denoting a valid directory (relative or absolute)
     */
	public static void setOutputDir(String outputdir)  {
		jasco.options.Options.setOutputDir(outputdir);
	}
	
	/**
     * gets the output directory where all entities generated by this class are placed 
	 *
     * @return a string denoting a valid directory (relative or absolute)
     */
	public static String getOutputDir()  {
		return jasco.options.Options.getOutputDir();
	}
	
	/**
     * returns the classpath represented as a string separated by ; on windows and : on linux
	 *
     * @return classpath string
     */
	public static String getClassPath()  {
		return jasco.options.Options.getClassPath();
	}
	
	/**
     * adds a single directory or jar file to the classpath
	 *
     * @param path	a path or jar file
     */
	public static void addToClassPath(String path)  {
		jasco.options.Options.addToClassPath(path);
	}
	
	/**
     * Compiles an aspect bean and places it in the output directory.
	 *
     * @param aAspect	a file string containing an aspect bean definition	
     */
	public static void compileAspect(String aAspect) throws FileNotFoundException
	{	
		checkFile(aAspect);  
		FileInputStream file = null;
		try 
		{
		 	file=new java.io.FileInputStream(aAspect);
			jasco.tools.jascoparser.JavaCharStream stream = new jasco.tools.jascoparser.JavaCharStream(file,1,1);
			JascoParserTokenManager tokenmanager = new JascoParserTokenManager(stream);
			jasco.tools.aspectparser.AspectParser parser = new jasco.tools.aspectparser.AspectParser(tokenmanager,aAspect);
			parser.start();
			jasco.tools.aspectparser.PAspect theaspect = parser.getParsedAspect();
			jasco.tools.aspectparser.AspectClassesGenerator generator = new jasco.tools.aspectparser.AspectClassesGenerator(theaspect);
			generator.generate();
		}
		catch (IOException e)
		{
			Logger.getInstance().showError(e.getMessage());
			
		}
		finally {
			try {
				file.close();
			}catch(IOException ex)  {
				Logger.getInstance().showError(ex.getMessage());
			
				}
		}
	}
	
	/**
     * Compiles a traversal connector and places it in the output directory.
	 *
     * @param connector	a string denoting a file containing a traversal connector definition	
     */
	public static void compileTraversal(String connector) throws FileNotFoundException
	{	
		checkFile(connector);
		FileInputStream file = null;
		try 
		{
			file=new java.io.FileInputStream(connector);
			JavaCharStream stream = new JavaCharStream(file,1,1);
			JascoParserTokenManager tokenmanager = new JascoParserTokenManager(stream);
			jasco.tools.traversalparser.TraversalParser parser = new jasco.tools.traversalparser.TraversalParser(tokenmanager,connector);
			parser.start();
			jasco.tools.traversalparser.TraversalSpec theaspect = parser.getTraversalSpec();
			jasco.tools.traversalparser.TraversalClassesGenerator generator = new jasco.tools.traversalparser.TraversalClassesGenerator(theaspect);
			generator.generate();
		}
		catch (IOException e)
		{
			Logger.getInstance().showError(e.getMessage());
		}
		finally {
			try {
				file.close();
			}catch(IOException ex)  {
				Logger.getInstance().showError(ex.getMessage());
			
				}
		}
	}
	
	/**
     * Tranforms a java bean and places the transformed version in the output directory
	 *
     * @param javaBean	a string denoting a fully qualified class name (e.g. java.lang.String)
	 * @param excluded	a list of method names that cannot be trapped
     */
	public static void transformBeanInOutputDir(String javaBean,Vector excluded) 
	{
		BeanTransformerJavassist bt = new BeanTransformerJavassist();
		bt.transformBean(javaBean,excluded);	
	}
	
	/**
     * Tranforms a java bean and places the transformed version in the output directory. 
	 * Runs in a different virtual machine.
	 *
     * @param javaBean	a string denoting a fully qualified class name (e.g. java.lang.String)
	 * @param excluded	a list of method names that cannot be trapped (not supported yet here)
     */
	public static void transformBeanInOutputDirInNewVM(String javaBean,Vector excluded) 
	{	
		try 
		{
		//String name = Options.getInstallPath()+File.separator+"temp"+File.separator+"transform.bat";
		String name = Options.getOutputDir()+File.separator+"transform.bat";
		FileWriter writer = new FileWriter(name);
		//System.out.println(name);
		//System.out.println("TransformBean "+javaBean+" \"-classpath:"+Options.getClassPath()+Options.getOutputDir()+"\" \"-outputdir:"+Options.getOutputDir()+"\""+JavaGenerator.NEWLINE);
		String od = Options.getOutputDir();
		if(od.endsWith("\\"))
			od=od.substring(0,od.length()-1);
		String cp=Options.getClassPath();
		if(cp.endsWith("\\"))
			cp=cp.substring(0,cp.length()-1);
		
		String transformCmd = "TransformBean";
		if(Options.getInstallPath()!=null&&!Options.getInstallPath().equals(""))
			transformCmd=Options.getInstallPath()+File.separator+"bin"+File.separator+transformCmd;
			
		writer.write(transformCmd+" "+javaBean+" \"-classpath:"+cp+od+"\" \"-outputdir:"+od+"\""+JavaGenerator.NEWLINE);
		writer.close();
		jasco.util.CommandRunner.runCommand(name);
		File file = new File(name);
		if(Options.deleteTempFiles())
			file.delete();
		}
		catch (IOException e)
		{
			Logger.getInstance().showError(e.getMessage());
		}
	}
	
	protected static void checkFile(String f)  throws FileNotFoundException
	{		
		if(!new File(f).exists())
			throw new FileNotFoundException(f);
	}
	
	/**
     * Compile a connector and places the result in the output directory. 
	 *
     * @param aConnector	a string denoting a file that contains a connector definition
	 * @return the name of the connector
     */
	public static String compileConnector(String aConnector) throws FileNotFoundException
	{	
		checkFile(aConnector);
		String name = null;
		FileInputStream file = null;
		try 
		{	
			file = new java.io.FileInputStream(aConnector);
			JavaCharStream stream = new JavaCharStream(file,1,1);
			JascoParserTokenManager tokenmanager = new JascoParserTokenManager(stream);
			jasco.tools.connectorparser.ConnectorParser parser = new jasco.tools.connectorparser.ConnectorParser(tokenmanager,aConnector);
			//parser.setIsCheckingClasses(false); testing 
			parser.start();
			jasco.tools.connectorparser.PConnector theconnector = parser.getParsedConnector();
			name = theconnector.getConnectorName();
			
			jasco.tools.connectorparser.ConnectorClassesGenerator generator = new jasco.tools.connectorparser.ConnectorClassesGenerator(theconnector);
		
			generator.generate();			
		} 
		catch (IOException e)
		{	
			Logger.getInstance().showError(e.getMessage());
		}
		finally {
			try {
				file.close();
			}catch(IOException ex)  {
				Logger.getInstance().showError(ex.getMessage());
			
				}
		}
		return name;
	}
	
	/**
	 * Load a connector so that all declared aspect beans are active. <br> Not necessary when JAsCo searches the classpath
	 * for new connectors and the outputdir is in the classpath.
	 *
     * @param aConnector	a string denoting a file that contains a connector definition
     */
	public static void loadConnector(jasco.runtime.connector.Connector aConnector)  throws Exception
	{
		loadConnector(aConnector,null);
	}
	
	private static void loadConnector(jasco.runtime.connector.Connector aConnector,String f)  throws Exception
	{
		Method m = aConnector.getClass().getMethod("load",new Class[]  {String.class});
		m.invoke(null,new Object[]  {f});
	}
	
	/**
	 * Unload a connector. <br> Not necessary when JAsCo searches the classpath
	 * for new connectors and the outputdir is in the classpath. Then the connector is auto unloaded when deleted.
	 *
     * @param aConnector	a connector singleton instance
     */
	public static void unloadConnector(jasco.runtime.connector.Connector aConnector) 
	{
		jasco.runtime.ConnectorRegistry.unRegisterConnector(aConnector);
	}
	
	
	/**
     * Compile a connector and places the result in the output directory. 
	 * The connector is also loaded by the JAsCo system and all declared aspect beans are activated.
	 * <br>
	 * Not necessary when JAsCo searches the classpath
	 * for new connectors.
	 *
	 * <p> Make sure outputdir is in the static startup classpath, fails otherwise!!!
	 *
     * @param aConnector	a string denoting a file that contains a connector definition
     */
	public static void compileAndLoadConnector(String aConnector)  throws Exception
	{
		String name = compileConnector(aConnector);
		if(name==null)
			throw new IllegalArgumentException("Compiling "+aConnector+" failed!");
		
		String classname = "Connector."+name;
		Class c = Class.forName(classname);
		Method m = c.getMethod("getConnector",new Class[0]);
		Connector  co = (Connector) m.invoke(null,new Object[0]);
		loadConnector(co,aConnector);
	}
	
	/**
     * Remove a connector from the output directory. 
	 *
     * @param aConnector	A connector NAME (not filename if different).
     */
	public static void removeConnectorInOutputDir(String aConnector) 
	{
	
			File file = new File(jasco.options.Options.getOutputDir(),jasco.options.Options.getConnectorPackage());				  		
			if(!aConnector.endsWith(".class"))
				aConnector=aConnector+".class";
			File cFile = new File(file,aConnector);

			if(cFile.exists())
				cFile.delete();
			else System.err.println("could not locate connector: "+aConnector);	
			
			
			
			jasco.runtime.ConnectorPersistance.removeConfFile(cFile);	
		
	}
	
	/**
     * Remove a connector from the classpath. Searches the complete classpath, if multiple occurances our found
	 * all of them are deleted.. 
	 *
     * @param aConnector	A connector NAME (not filename if different).
     */
	public static void removeConnectorInClassPath(String aConnector) 
	{	
		if(!aConnector.endsWith(".class"))
			aConnector=aConnector+".class";
	
		String classpath = System.getProperty("java.class.path");
		StringTokenizer tokenizer = new StringTokenizer(classpath,";");
		while (tokenizer.hasMoreTokens()) 
		{
			String token = tokenizer.nextToken();
			File file = new File(token);
			File connectorpath = new File(file,"Connector");
			if (connectorpath.exists() && connectorpath.isDirectory()) 
			{	
				
				
				File f = new File(connectorpath,aConnector);
				if (f.exists())  {
					jasco.runtime.ConnectorPersistance.removeConfFile(f);
					f.delete();
					
				}
					
			}
		}
	}
	
	/**
     * Forces the connector registry to load new connectors NOW.
	 *
     */
	public static void loadConnectors()  {
		jasco.runtime.ConnectorRegistry.forceConnectorLoad();
	}
	
	/**
     * Alters the connector loading priority to a new one. Halts the JAsCo aspect framework and restarts it. 
	 * Only use this if absolutely required. You can use the -Djasco.connector.loadpriority startup switch for setting the priority at startup.
	 *
	 * @param pri new Priority has to be between [1..10]
     */
	public static void modifyConnectorLoadingPriority(int pri)  {
	
		jasco.options.Options.modifyConnectorLoadingPriority(pri);
	}
	
	/**
     * Creates an introspector gui window 
	 *
     */
	public static void createIntrospectorWindow()  {
		
		new jasco.tools.gui.ConnectorRegistryIntrospector();
	}
	
	private static void startiep(String s) {
		System.out.println("string");
	}
	
	private static void startiep(Object s) {
		System.out.println("object");
		
	}
	
	/**
     * testing method, do not use!
	 * 
     */  
	public static void main(String[] args)  throws Exception {
		
		java.util.List l = new java.util.ArrayList();
		
		Object s = new String("thestring");  
		startiep(s);
		
		//Options.addToClassPath("e:/programs/JAsCo/examples/MISC/Generated");
		
		/*
		Options.setOutputDir("e:/programs/JAsCo/examples/MISC/Generated");
		
		Jasco.compileAndLoadConnector("e:/programs/JAsCo/examples/MISC/basic.con");
		Jasco.transformBeanInOutputDir("javax.swing.MyJButton",new Vector());
		
		Class button = Class.forName("javax.swing.MyJButton");
		Object b = button.newInstance();
		Method m = button.getMethod("addActionListener", new Class[] {Class.forName("java.awt.event.ActionListener")});
		Object res = m.invoke(b,new Object[] {null});
		//System.out.println(res);
		*/
		
		Object [] test = new Object[0];
		Object [][] test2 = new Object[0][0];
		int [][] test3 = new int[0][0];
		short [][] test4 = new short[0][0];
		double [][] test5 = new double[0][0];
		float [][] test10 = new float[0][0];
		byte [][] test6 = new byte[0][0];
		char [][] test7 = new char[0][0];
		boolean [][] test8 = new boolean[0][0];
		long [][] test9 = new long[0][0];
		System.out.println(test.getClass().getName());
		System.out.println(test2.getClass().getName());
		System.out.println(test3.getClass().getName());
		System.out.println(test4.getClass().getName());
		System.out.println(test5.getClass().getName());
		System.out.println(test6.getClass().getName());
		System.out.println(test7.getClass().getName());
		System.out.println(test8.getClass().getName());
		System.out.println(test9.getClass().getName());
		System.out.println(test10.getClass().getName());
		
		System.out.println(Object.class.getName());
		/*
		Options.addSinglePathToClassPath("E:\\programs\\JAsCo\\examples\\MISC\\Generated");
		Options.loadClass("javax.swing.MyJButton");
		
		//Options.addSinglePathToClassPath("E:\\programs\\JAsCo\\examples\\MISC\\Generated");
		Options.loadClass("javax.swing.MyJButton");
		
		/*
		Options.showDebugOutput(true);
		Options.setInstallPath("e:/programs/JAsCo");
		//Options.setConnectorLoadInterval(0);
		Options.setConnectorLoadPath("e:/programs/JAsCo");
		Options.addToClassPath("e:/programs/JAsCo/examples/MISC/generated\\");
		Options.setOutputDir("e:/programs/JAsCo/examples/MISC/generated\\");
		//Jasco.transformBeanInOutputDirInNewVM("javax.swing.MyJButton",new Vector());
		System.out.println("boe");
		
		//Jasco.compileConnector("e:/programs/JAsCo/examples/MISC/basic.con");
		
		loadConnectors();
		
		Thread.sleep(10000);
		
		
		modifyConnectorLoadingPriority(10);
		
		System.out.println("yte");
		/*
		Object [] test = new Object[3];
		
		System.out.println(test.getClass().getName());
		
		//Object[] args2 = new Object[Integer.MAX_VALUE];
		
		for(int i=0;i<100;++i)  
			bench();*/
	}
}